import logging
import json
from abc import ABC, abstractmethod
import boto3
from botocore.exceptions import ClientError

from redis import Redis

log = logging.getLogger(__name__)


class AbstractBackend(ABC):
    @abstractmethod
    def push(self, queue_name, data):
        """Push a message onto the queue.

        :param data: The Python native object to be pushed onto the queue.
        """
        pass

    @abstractmethod
    def listen(self, queue_names):
        """Iterator over messages in the queue.

        :param queue_names: The names of the queues to listen on.

        :returns: A triple consisting of a message handle, the queue name and
                  the structured message (likely a Python dict).
        """
        pass

    @abstractmethod
    def sign_off(self, handle):
        """Notify the backend that message is dealt with appropriately.  The
        backend should ensure that the message is not picked up again.

        :param handle: The backend-specific message handle.  Should the first
            item in the triples generated by :meth:`listen`.
        """
        pass


class RedisBackend(AbstractBackend):
    def __init__(self, redis):
        self._redis = redis

    def push(self, queue_name, data):
        self._redis.rpush(queue_name, json.dumps(data))

    def listen(self, queue_names):
        while True:
            packed = self._redis.blpop(queue_names, 30)
            if not packed:
                continue
            data = json.loads(packed[1].decode("utf-8"))
            queue_name = packed[0].decode("utf-8")
            # Sign-off functionality for redis is a noop, hence a message
            # handle is not needed: hence None
            yield (None, queue_name, data)

    def sign_off(self, handle):
        pass

    @classmethod
    def from_config(cls, config):
        if "type" in config:
            config = config.copy()
            del config["type"]
        return cls(Redis(**config))

    def __repr__(self):
        return "<{}(redis={})>".format(self.__class__.__name__, self._redis)


class SqsBackend(AbstractBackend):
    """SQS backend. Currently only supports listening to a single queue."""

    # Set long polling delay to maximum duration (20s)
    _WAIT_TIME_SECONDS = 20

    def __init__(self, client):
        self._client = client
        log.info("Created backend %s", self)

    def _get_queue_url(self, queue_name):
        try:
            response = self._client.get_queue_url(QueueName=queue_name)
        except ClientError:
            self._client.create_queue(QueueName=queue_name)
        response = self._client.get_queue_url(QueueName=queue_name)
        return response.get("QueueUrl")

    def push(self, queue_name, data):
        queue_url = self._get_queue_url(queue_name)
        self._client.send_message(QueueUrl=queue_url, MessageBody=json.dumps(data))

    def listen(self, queue_names):
        queue_url = self._get_queue_url(queue_names[0])
        while True:
            messages = self._client.receive_message(
                QueueUrl=queue_url, WaitTimeSeconds=self._WAIT_TIME_SECONDS
            )
            if "Messages" not in messages:
                log.debug("Timeout waiting for messages")
                continue
            for mess in messages["Messages"]:
                data = json.loads(mess["Body"])
                handle = {"ReceiptHandle": mess["ReceiptHandle"], "QueueUrl": queue_url}
                yield (handle, queue_names[0], data)

    def sign_off(self, handle):
        self._client.delete_message(**handle)

    @classmethod
    def from_config(cls, _config):
        return cls(boto3.client("sqs"))

    def __repr__(self):
        return "<{}(client={})>".format(self.__class__.__name__, self._client)



from totec.client import JobClient
from totec.store import InMemoryStore
from totec.backend import RedisBackend, SqsBackend
from totec.listener import JobListener


def new_backend(config):
    ty = config.get("type")
    if ty == "redis":
        return RedisBackend.from_config(config)
    if ty == "sqs":
        return SqsBackend.from_config(config)
    raise ValueError("No such backend {}".format(ty))


def new_store(config):
    ty = config.get("type")
    if ty == "in-memory":
        return InMemoryStore()
    raise ValueError("No such store {}".format(ty))


def new_job_client(config, store=None, backend=None):
    url_template = config["url_template"]
    queue_template = config["queue_template"]
    store = store or new_store(config["store"])
    backend = backend or new_backend(config["backend"])
    return JobClient(store, backend, url_template, queue_template)


def new_job_listener(config):
    queue_names = config["queue_names"]
    outputs_encoder = config["outputs_encoder"]
    backend = new_backend(config["backend"])
    return JobListener(backend, queue_names, outputs_encoder)



from abc import ABC


class AbstractStore(ABC):
    """
    Interface for storing jobs that is used by :class:`JobClient` to keep track
    which jobs are outstanding.
    """

    def put(self, job):
        pass

    def get(self, tag):
        pass

    def list(self):
        pass


class InMemoryStore(AbstractStore):
    """Implementation that stores the jobs in memory.
    """

    def __init__(self):
        self._store = {}

    def put(self, job):
        self._store[job.tag] = job

    def get(self, tag):
        return self._store.get(tag)

    def list(self):
        for job in self._store.values():
            yield job



from datetime import datetime as dt
import logging
import uuid
import enum

log = logging.getLogger(__name__)


class SimpleJob:
    """Class to describe a job.  From an instance of this class an appropriate
    worker should be able to infer which computation (i.e., which job, and which
    inputs) it should run to get the outputs.
    """

    def __init__(self, job_type, tag, status, inputs):
        """Create an instance.

        :param job_type: An identifier string indicating the type of the job to
            create.  This identifier is shared between the creator of the job
            and the workers.
        :param tag: The identifier string for the job.  Should be unique for
            jobs created at the same :class:`JobClient`.
        :param status: The initial :class:`JobStatus` of the job
        :param inputs: The inputs for the job.  It is up to the client that the
            inputs are correct for the job type.
        """
        self.tag = tag
        self.transitions = []
        self.inputs = inputs
        self.job_type = job_type
        self.status = status
        self.outputs = None

    @property
    def status(self):
        """Return the status."""
        return self.transitions[-1][0]

    @status.setter
    def status(self, status):
        """Set the status of the job.

        :param status: The new status, should be an instance of
            :class:`JobStatus`

        :raises ValueError: If the transition from the old to the new status
            is not allowed.
        """
        self.transitions.append((status, dt.utcnow()))

    @classmethod
    def create(cls, job_type, inputs):
        """Create a new job, with initial status :class:`JobStatus.CREATED`.

        See :meth:`SimpleJob.__init__`
        """
        tag = str(uuid.uuid4())
        return cls(job_type, tag, JobStatus.CREATED, inputs)

    def to_dict(self):
        """Create a JSON serializable representation of the job."""
        obj = {
            "inputs": self.inputs,
            "type": self.job_type,
            "tag": self.tag,
            "transitions": [
                (status.name, t.isoformat()) for (status, t) in self.transitions
            ],
        }
        if self.outputs is not None:
            obj["outputs"] = self.outputs
        return obj

    def transition(self, status, data=None):
        """Update the job's status.

        :param status: The new status of the job.
        :param data: The data involved in the state transition, if any.  This is
            usually either the outputs of the job for status
            :class:`JobStatus.SUCCES`, or a description of the failure for a
            failure state such as :class:`JobStatus.FAILED`.

        :returns: None
        """

        if (self.status, status) in _ALLOWED_TRANSITIONS:
            log.info(
                "Transitioning job %s from %s to %s", self.tag, self.status, status
            )
            self.status = status
            self.outputs = data
        else:
            raise ValueError(
                "Cannot transition from {} to {}".format(self.status, status)
            )

    def __repr__(self):
        return "<{}(tag={}, status={})>".format(
            self.__class__.__name__, self.tag, self.status
        )


class JobStatus(enum.Enum):
    CREATED = 0x01
    SUBMITTED = 0x02
    SUBMIT_FAILED = 0x04
    SUCCES = 0x08
    FAILED = 0x10

    def __repr__(self):
        return self.name

    @property
    def code(self):
        return self.name

    def __str__(self):
        return repr(self)

    @classmethod
    def from_code(cls, s):
        return cls.from_string(s)

    @classmethod
    def from_string(cls, s):
        for value in cls:
            if value.name == s:
                return value
        raise ValueError("{} not a valid {}".format(s, cls.__name__))


# yapf: disable
_ALLOWED_TRANSITIONS = \
    [(JobStatus.CREATED, JobStatus.SUBMITTED),
     (JobStatus.SUBMITTED, JobStatus.SUBMIT_FAILED),
     (JobStatus.SUBMITTED, JobStatus.SUCCES),
     (JobStatus.SUBMITTED, JobStatus.FAILED)]
# yapf: enable
